@@ -3,9 +3,10 @@ import dataclasses
 import json
 import sys
 from pathlib import Path
-from typing import Any, Callable, Dict, List, Type, TypeVar, Union
+from typing import Any, Callable, Dict, List, Type, TypeVar, Union  # noqa: TID251
 
 from .._compat.builtins import ExceptionGroup
+from .._compat.typing import get_args, get_origin
 from .model.cache import Cache
 from .model.cmakefiles import CMakeFiles
 from .model.codemodel import CodeModel, Target
@@ -39,7 +40,7 @@ class Converter:
         return self.make_class(data, Index)
 
     def _load_from_json(self, name: Path, target: Type[T]) -> T:
-        with open(self.base_dir / name, encoding="utf-8") as f:
+        with self.base_dir.joinpath(name).open(encoding="utf-8") as f:
             data = json.load(f)
 
         return self.make_class(data, target)
@@ -58,7 +59,8 @@ class Converter:
         input_dict = {}
         exceptions: List[Exception] = []
 
-        for field in dataclasses.fields(target):
+        # We don't have DataclassInstance exposed in typing yet
+        for field in dataclasses.fields(target):  # type: ignore[arg-type]
             json_field = field.name.replace("_v", "-v").replace(
                 "cmakefiles", "cmakeFiles"
             )
@@ -70,7 +72,7 @@ class Converter:
                 except TypeError as err:
                     msg = f"Failed to convert field {field.name!r} of type {field.type}"
                     if sys.version_info < (3, 11):
-                        err.__notes__ = getattr(err, "__notes__", []) + [msg]  # type: ignore[attr-defined]
+                        err.__notes__ = [*getattr(err, "__notes__", []), msg]  # type: ignore[attr-defined]
                     else:
                         err.add_note(msg)  # pylint: disable=no-member
                     exceptions.append(err)
@@ -78,18 +80,21 @@ class Converter:
                     exceptions.append(err)
 
         if exceptions:
-            raise ExceptionGroup(f"Failed converting {target}", exceptions)
+            msg = f"Failed converting {target}"
+            raise ExceptionGroup(msg, exceptions)
 
         return target(**input_dict)
 
     def _convert_any(self, item: Any, target: Type[T]) -> T:
         if dataclasses.is_dataclass(target):
-            return self.make_class(item, target)
-        if hasattr(target, "__origin__"):
-            if target.__origin__ == list:  # type: ignore[attr-defined]
-                return [self._convert_any(i, target.__args__[0]) for i in item]  # type: ignore[return-value,attr-defined]
-            if target.__origin__ == Union:  # type: ignore[attr-defined]
-                return self._convert_any(item, target.__args__[0])  # type: ignore[no-any-return,attr-defined]
+            # We don't have DataclassInstance exposed in typing yet
+            return self.make_class(item, target)  # type: ignore[return-value]
+        origin = get_origin(target)
+        if origin is not None:
+            if origin == list:
+                return [self._convert_any(i, get_args(target)[0]) for i in item]  # type: ignore[return-value]
+            if origin == Union:
+                return self._convert_any(item, get_args(target)[0])  # type: ignore[no-any-return]
 
         return target(item)  # type: ignore[call-arg]
 
