@@ -1,12 +1,61 @@
+"""
+This is the configuration tooling for scikit-build-core. This is build around
+the :class:`Source` Protocol. Sources are created with some input (like a toml
+file for the :class:`TOMLSource`). Sources also usually have some prefix (like
+``tool.scikit-build``) as well. The :class:`SourceChain` holds a collection of
+Sources, and is the primary way to use them.
+
+An end user interacts with :class:`SourceChain` via ``.convert_target``, which
+takes a Dataclass class and returns an instance with fields populated.
+
+Example of usage::
+
+    sources = SourceChain(TOMLSource("tool", "mypackage", settings=pyproject_dict), ...)
+    settings = sources.convert_target(SomeSettingsModel)
+
+    unrecognized_options = list(source.unrecognized_options(SomeSettingsModel)
+
+
+Naming conventions:
+
+- ``model`` is the complete Dataclass.
+- ``target`` is the type to convert a single item to.
+- ``settings`` is the input data source (unless it already has a name, like
+  ``env``).
+- ``options`` are the names of the items in the ``model``, formatted in the
+  style of the current Source.
+- ``fields`` are the tuple of strings describing a nested field in the
+  ``model``.
+
+When setting up your dataclasses, these types are handled:
+
+- ``str``: A string type, nothing special.
+- ``bool``: Supports bool in TOML, not handled in envvar/config (so only useful in a Union)
+- Any callable (`Path`, `Version`): Passed the string input.
+- ``Optional[T]``: Treated like T. Default should be None, since no input format supports None's.
+- ``Union[str, ...]``: Supports other input types in TOML form (bool currently). Otherwise a string.
+- ``List[T]``: A list of items. `;` separated supported in EnvVar/config forms. T can be a dataclass (TOML only).
+- ``Dict[str, T]``: A table of items. TOML supports a layer of nesting. Any is supported as an item type.
+- ``Literal[...]``: A list of strings, the result must be in the list.
+
+These are supported for JSON schema generation for the TOML, as well.
+
+Integers/floats would be easy to add, but haven't been needed yet.
+"""
+
+
 from __future__ import annotations
 
 import dataclasses
 import os
-from collections.abc import Generator, Iterator, Mapping, Sequence
+import typing
 from typing import Any, TypeVar, Union
 
 from .._compat.builtins import ExceptionGroup
-from .._compat.typing import Protocol, runtime_checkable
+from .._compat.typing import Literal, Protocol, get_args, get_origin
+
+if typing.TYPE_CHECKING:
+    from collections.abc import Generator, Iterator, Mapping, Sequence
 
 T = TypeVar("T")
 
@@ -17,98 +66,80 @@ def __dir__() -> list[str]:
     return __all__
 
 
-def _dig_strict(dict_: Mapping[str, Any], *names: str) -> Any:
+def _dig_strict(__dict: Mapping[str, Any], *names: str) -> Any:
     for name in names:
-        dict_ = dict_[name]
-    return dict_
+        __dict = __dict[name]
+    return __dict
 
 
-def _dig_not_strict(dict_: Mapping[str, Any], *names: str) -> Any:
+def _dig_not_strict(__dict: Mapping[str, Any], *names: str) -> Any:
     for name in names:
-        dict_ = dict_.get(name, {})
-    return dict_
+        __dict = __dict.get(name, {})
+    return __dict
 
 
-def _dig_fields(__opt: object, *names: str) -> Any:
+def _dig_fields(__opt: Any, *names: str) -> Any:
     for name in names:
         fields = dataclasses.fields(__opt)
         types = [x.type for x in fields if x.name == name]
         if len(types) != 1:
-            raise KeyError("Could not access {'.'.join(names)}")
+            msg = f"Could not access {'.'.join(names)}"
+            raise KeyError(msg)
         (__opt,) = types
     return __opt
 
 
-@runtime_checkable
-class TypeLike(Protocol):
-    @property
-    def __origin__(self) -> Any:
-        ...
-
-    @property
-    def __args__(self) -> list[Any]:
-        ...
-
-
 def _process_union(target: type[Any]) -> Any:
     """
-    Selects the non-None item in an Optional or Optional-like Union. Passes through non-Unions.
+    Filters None out of Unions. If a Union only has one item, return that item.
     """
 
-    if (
-        not isinstance(target, TypeLike)
-        or not hasattr(target, "__origin__")
-        or target.__origin__ is not Union
-    ):
-        return target
+    origin = get_origin(target)
 
-    if len(target.__args__) == 2:
-        items = list(target.__args__)
-        if type(None) not in items:
-            raise AssertionError(f"None must be in union, got {items}")
-        items.remove(type(None))
-        return items[0]
+    if origin is Union:
+        non_none_args = [a for a in get_args(target) if a is not type(None)]
+        if len(non_none_args) == 1:
+            return non_none_args[0]
+        return Union[tuple(non_none_args)]
 
-    raise AssertionError("Only Unions with None supported")
-
-
-def _get_target_raw_type(target: type[Any]) -> type[Any]:
-    """
-    Takes a type like Optional[str] and returns str,
-    or Optional[Dict[str, int]] and returns dict.
-    """
-
-    target = _process_union(target)
-    # The hasattr is required for Python 3.7, though not quite sure why
-    if isinstance(target, TypeLike) and hasattr(target, "__origin__"):
-        return target.__origin__
     return target
 
 
-def _get_inner_type(target: type[Any]) -> type[Any]:
+def _get_target_raw_type(target: type[Any]) -> Any:
     """
-    Takes a types like List[str] and returns str,
-    or Dict[str, int] and returns int.
+    Takes a type like ``Optional[str]`` and returns str, or ``Optional[Dict[str,
+    int]]`` and returns dict. Returns Union for a Union with more than one
+    non-none type. Literal is also a valid return.
     """
 
-    raw_target = _get_target_raw_type(target)
     target = _process_union(target)
+    origin = get_origin(target)
+    return origin or target
+
+
+def _get_inner_type(__target: type[Any]) -> type[Any]:
+    """
+    Takes a types like ``List[str]`` and returns str,
+    or ``Dict[str, int]`` and returns int.
+    """
+
+    raw_target = _get_target_raw_type(__target)
+    target = _process_union(__target)
     if raw_target == list:
-        assert isinstance(target, TypeLike)
-        return target.__args__[0]
+        return get_args(target)[0]  # type: ignore[no-any-return]
     if raw_target == dict:
-        assert isinstance(target, TypeLike)
-        return target.__args__[1]
-    raise AssertionError("Expected a list or dict")
+        return get_args(target)[1]  # type: ignore[no-any-return]
+    msg = f"Expected a list or dict, got {target!r}"
+    raise AssertionError(msg)
 
 
-def _nested_dataclass_to_names(target: type[Any], *inner: str) -> Iterator[list[str]]:
+def _nested_dataclass_to_names(__target: type[Any], *inner: str) -> Iterator[list[str]]:
     """
-    Yields each entry, like ("a", "b", "c") for a.b.c
+    Yields each entry, like ``("a", "b", "c")`` for ``a.b.c``.
     """
 
-    if dataclasses.is_dataclass(target):
-        for field in dataclasses.fields(target):
+    if dataclasses.is_dataclass(__target):
+        for field in dataclasses.fields(__target):
             yield from _nested_dataclass_to_names(field.type, *inner, field.name)
     else:
         yield list(inner)
@@ -117,23 +148,40 @@ def _nested_dataclass_to_names(target: type[Any], *inner: str) -> Iterator[list[
 class Source(Protocol):
     def has_item(self, *fields: str, is_dict: bool) -> bool:
         """
-        Check if the source contains a chain of fields. For example, feilds =
-        [Field(name="a"), Field(name="b")] will check if the source contains the
-        key "a.b".
+        Check if the source contains a chain of fields. For example, ``fields =
+        [Field(name="a"), Field(name="b")]`` will check if the source contains the
+        key "a.b". ``is_dict`` should be set if it can be nested.
         """
         ...
 
     def get_item(self, *fields: str, is_dict: bool) -> Any:
+        """
+        Select an item from a chain of fields. Raises KeyError if
+        the there is no item. ``is_dict`` should be set if it can be nested.
+        """
         ...
 
     @classmethod
     def convert(cls, item: Any, target: type[Any]) -> object:
+        """
+        Convert an ``item`` from the base representation of the source's source
+        into a ``target`` type. Raises TypeError if the conversion fails.
+        """
         ...
 
     def unrecognized_options(self, options: object) -> Generator[str, None, None]:
+        """
+        Given a model, produce an iterator of all unrecognized option names.
+        Empty iterator if this can't be computed for the source (like for
+        environment variables).
+        """
         ...
 
     def all_option_names(self, target: type[Any]) -> Iterator[str]:
+        """
+        Given a model, produce a list of all possible names (used for producing
+        suggestions).
+        """
         ...
 
 
@@ -150,21 +198,25 @@ class EnvSource:
         names = [field.upper() for field in fields]
         return "_".join([self.prefix, *names] if self.prefix else names)
 
-    # pylint: disable-next=unused-argument
-    def has_item(self, *fields: str, is_dict: bool) -> bool:
+    def has_item(self, *fields: str, is_dict: bool) -> bool:  # noqa: ARG002
         name = self._get_name(*fields)
-        return name in self.env and self.env[name] != ""
+        return bool(self.env.get(name, ""))
 
-    # pylint: disable-next=unused-argument
-    def get_item(self, *fields: str, is_dict: bool) -> str | dict[str, str]:
+    def get_item(
+        self, *fields: str, is_dict: bool  # noqa: ARG002
+    ) -> str | dict[str, str]:
         name = self._get_name(*fields)
         if name in self.env:
             return self.env[name]
-        raise KeyError(f"{name!r} not found in environment")
+        msg = f"{name!r} not found in environment"
+        raise KeyError(msg)
 
     @classmethod
     def convert(cls, item: str, target: type[Any]) -> object:
         raw_target = _get_target_raw_type(target)
+        if dataclasses.is_dataclass(raw_target):
+            msg = f"Array of dataclasses are not supported in configuration settings ({raw_target})"
+            raise TypeError(msg)
         if raw_target == list:
             return [
                 cls.convert(i.strip(), _get_inner_type(target)) for i in item.split(";")
@@ -174,15 +226,25 @@ class EnvSource:
             return {k: cls.convert(v, _get_inner_type(target)) for k, v in items}
 
         if raw_target is bool:
-            result = item.strip().lower() not in {"0", "false", "off", "no", ""}
-            return result
+            return item.strip().lower() not in {"0", "false", "off", "no", ""}
+
+        if raw_target is Union and str in get_args(target):
+            return item
+
+        if raw_target is Literal:
+            if item not in get_args(_process_union(target)):
+                msg = f"{item!r} not in {get_args(_process_union(target))!r}"
+                raise TypeError(msg)
+            return item
 
         if callable(raw_target):
             return raw_target(item)
-        raise AssertionError(f"Can't convert target {target}")
+        msg = f"Can't convert target {target}"
+        raise TypeError(msg)
 
-    # pylint: disable-next=unused-argument
-    def unrecognized_options(self, options: object) -> Generator[str, None, None]:
+    def unrecognized_options(
+        self, options: object  # noqa: ARG002
+    ) -> Generator[str, None, None]:
         yield from ()
 
     def all_option_names(self, target: type[Any]) -> Iterator[str]:
@@ -192,9 +254,10 @@ class EnvSource:
 
 
 def _unrecognized_dict(
-    settings: Mapping[str, Any], options: object, above: Sequence[str]
+    settings: Mapping[str, Any], options: Any, above: Sequence[str]
 ) -> Generator[str, None, None]:
     for keystr in settings:
+        # We don't have DataclassInstance exposed in typing yet
         matches = [
             x for x in dataclasses.fields(options) if x.name.replace("_", "-") == keystr
         ]
@@ -223,7 +286,7 @@ class ConfSource:
         *prefixes: str,
         settings: Mapping[str, str | list[str]],
         verify: bool = True,
-    ):
+    ) -> None:
         self.prefixes = prefixes
         self.settings = settings
         self.verify = verify
@@ -252,36 +315,50 @@ class ConfSource:
             }
             if d:
                 return d
-            raise KeyError(f"Dict items {name}.* not found in settings")
+            msg = f"Dict items {name}.* not found in settings"
+            raise KeyError(msg)
         if name in self.settings:
             return self.settings[name]
 
-        raise KeyError(f"{name!r} not found in configuration settings")
+        msg = f"{name!r} not found in configuration settings"
+        raise KeyError(msg)
 
     @classmethod
     def convert(
         cls, item: str | list[str] | dict[str, str], target: type[Any]
     ) -> object:
         raw_target = _get_target_raw_type(target)
+        if dataclasses.is_dataclass(raw_target):
+            msg = f"Array of dataclasses are not supported in configuration settings ({raw_target})"
+            raise TypeError(msg)
         if raw_target == list:
             if isinstance(item, list):
                 return [cls.convert(i, _get_inner_type(target)) for i in item]
-            assert not isinstance(item, dict)
+            if isinstance(item, dict):
+                msg = f"Expected {target}, got {type(item).__name__}"
+                raise TypeError(msg)
             return [
                 cls.convert(i.strip(), _get_inner_type(target)) for i in item.split(";")
             ]
         if raw_target == dict:
             assert not isinstance(item, (str, list))
             return {k: cls.convert(v, _get_inner_type(target)) for k, v in item.items()}
-        assert not isinstance(
-            item, (list, dict)
-        ), "Can't convert list or dict to non-list/dict"
+        if isinstance(item, (list, dict)):
+            msg = f"Expected {target}, got {type(item).__name__}"
+            raise TypeError(msg)
         if raw_target is bool:
-            result = item.strip().lower() not in {"0", "false", "off", "no", ""}
-            return result
+            return item.strip().lower() not in {"0", "false", "off", "no", ""}
+        if raw_target is Union and str in get_args(target):
+            return item
+        if raw_target is Literal:
+            if item not in get_args(_process_union(target)):
+                msg = f"{item!r} not in {get_args(_process_union(target))!r}"
+                raise TypeError(msg)
+            return item
         if callable(raw_target):
             return raw_target(item)
-        raise AssertionError(f"Can't convert target {target}")
+        msg = f"Can't convert target {target}"
+        raise TypeError(msg)
 
     def unrecognized_options(self, options: object) -> Generator[str, None, None]:
         if not self.verify:
@@ -309,40 +386,64 @@ class ConfSource:
 
 
 class TOMLSource:
-    def __init__(self, *prefixes: str, settings: Mapping[str, Any]):
+    def __init__(self, *prefixes: str, settings: Mapping[str, Any]) -> None:
         self.prefixes = prefixes
         self.settings = _dig_not_strict(settings, *prefixes)
 
     def _get_name(self, *fields: str) -> list[str]:
         return [field.replace("_", "-") for field in fields]
 
-    # pylint: disable-next=unused-argument
-    def has_item(self, *fields: str, is_dict: bool) -> bool:
+    def has_item(self, *fields: str, is_dict: bool) -> bool:  # noqa: ARG002
         names = self._get_name(*fields)
         try:
             _dig_strict(self.settings, *names)
-            return True
         except KeyError:
             return False
+        return True
 
-    # pylint: disable-next=unused-argument
-    def get_item(self, *fields: str, is_dict: bool) -> Any:
+    def get_item(self, *fields: str, is_dict: bool) -> Any:  # noqa: ARG002
         names = self._get_name(*fields)
         try:
             return _dig_strict(self.settings, *names)
         except KeyError:
-            raise KeyError(f"{names!r} not found in configuration settings") from None
+            msg = f"{names!r} not found in configuration settings"
+            raise KeyError(msg) from None
 
     @classmethod
     def convert(cls, item: Any, target: type[Any]) -> object:
         raw_target = _get_target_raw_type(target)
-        if raw_target == list:
+        if dataclasses.is_dataclass(raw_target):
+            fields = dataclasses.fields(raw_target)
+            values = ((k.replace("-", "_"), v) for k, v in item.items())
+            return raw_target(
+                **{
+                    k: cls.convert(v, *[f.type for f in fields if f.name == k])
+                    for k, v in values
+                }
+            )
+        if raw_target is list:
+            if not isinstance(item, list):
+                msg = f"Expected {target}, got {type(item).__name__}"
+                raise TypeError(msg)
             return [cls.convert(it, _get_inner_type(target)) for it in item]
-        if raw_target == dict:
+        if raw_target is dict:
+            if not isinstance(item, dict):
+                msg = f"Expected {target}, got {type(item).__name__}"
+                raise TypeError(msg)
             return {k: cls.convert(v, _get_inner_type(target)) for k, v in item.items()}
+        if raw_target is Any:
+            return item
+        if raw_target is Union and type(item) in get_args(target):
+            return item
+        if raw_target is Literal:
+            if item not in get_args(_process_union(target)):
+                msg = f"{item!r} not in {get_args(_process_union(target))!r}"
+                raise TypeError(msg)
+            return item
         if callable(raw_target):
             return raw_target(item)
-        raise AssertionError(f"Can't convert target {target}")
+        msg = f"Can't convert target {target}"
+        raise TypeError(msg)
 
     def unrecognized_options(self, options: object) -> Generator[str, None, None]:
         yield from _unrecognized_dict(self.settings, options, self.prefixes)
@@ -354,40 +455,45 @@ class TOMLSource:
 
 
 class SourceChain:
-    def __init__(self, *sources: Source) -> None:
+    def __init__(self, *sources: Source, prefixes: Sequence[str] = ()) -> None:
+        """
+        Combine a collection of sources into a single object that can run
+        ``convert_target(dataclass)``.  An optional list of prefixes can be
+        given that will be prepended (dot separated) to error messages.
+        """
         self.sources = sources
+        self.prefixes = prefixes
 
     def __getitem__(self, index: int) -> Source:
         return self.sources[index]
 
     def has_item(self, *fields: str, is_dict: bool) -> bool:
-        for source in self.sources:
-            if source.has_item(*fields, is_dict=is_dict):
-                return True
-        return False
+        return any(source.has_item(*fields, is_dict=is_dict) for source in self.sources)
 
     def get_item(self, *fields: str, is_dict: bool) -> Any:
         for source in self.sources:
             if source.has_item(*fields, is_dict=is_dict):
                 return source.get_item(*fields, is_dict=is_dict)
-        raise KeyError(f"{fields!r} not found in any source")
-
-    @classmethod
-    def convert(cls, item: Any, target: type[T]) -> T:
-        raise NotImplementedError(
-            "SourceChain cannot convert items, use the result from has_item"
-        )
+        msg = f"{fields!r} not found in any source"
+        raise KeyError(msg)
 
     def convert_target(self, target: type[T], *prefixes: str) -> T:
+        """
+        Given a dataclass type, create an object of that dataclass filled
+        with the values in the sources.
+        """
+
         errors = []
-        prep = {}
-        for field in dataclasses.fields(target):
+        prep: dict[str, Any] = {}
+        for field in dataclasses.fields(target):  # type: ignore[arg-type]
             if dataclasses.is_dataclass(field.type):
                 try:
                     prep[field.name] = self.convert_target(
                         field.type, *prefixes, field.name
                     )
                 except Exception as e:
+                    name = ".".join([*self.prefixes, *prefixes, field.name])
+                    e.__notes__ = [*getattr(e, "__notes__", []), f"Field: {name}"]  # type: ignore[attr-defined]
                     errors.append(e)
                 continue
 
@@ -399,6 +505,8 @@ class SourceChain:
                     try:
                         tmp = source.convert(simple, field.type)
                     except Exception as e:
+                        name = ".".join([*self.prefixes, *prefixes, field.name])
+                        e.__notes__ = [*getattr(e, "__notes__", []), f"Field {name}"]  # type: ignore[attr-defined]
                         errors.append(e)
                         prep[field.name] = None
                         break
@@ -424,8 +532,9 @@ class SourceChain:
             errors.append(ValueError(f"Missing value for {field.name!r}"))
 
         if errors:
-            prefix_str = ".".join(prefixes)
-            raise ExceptionGroup(f"Failed converting {prefix_str}", errors)
+            prefix_str = ".".join([*self.prefixes, *prefixes])
+            msg = f"Failed converting {prefix_str}"
+            raise ExceptionGroup(msg, errors)
 
         return target(**prep)
 
@@ -433,6 +542,8 @@ class SourceChain:
         for source in self.sources:
             yield from source.unrecognized_options(options)
 
-    def all_option_names(self, target: type[Any]) -> Iterator[str]:
-        for source in self.sources:
-            yield from source.all_option_names(target)
+
+if typing.TYPE_CHECKING:
+    _: Source = typing.cast(EnvSource, None)
+    _ = typing.cast(ConfSource, None)
+    _ = typing.cast(TOMLSource, None)
