@@ -2,7 +2,7 @@ from __future__ import annotations
 
 import os
 import shutil
-import sys
+import sysconfig
 from pathlib import Path
 
 import pytest
@@ -24,7 +24,7 @@ has_ninja = shutil.which("ninja") is not None
 def prepare_env_or_skip() -> None:
     if (
         "CMAKE_GENERATOR" not in os.environ
-        and not sys.platform.startswith("win32")
+        and not sysconfig.get_platform().startswith("win")
         and not has_make
     ):
         if has_ninja:
@@ -33,9 +33,8 @@ def prepare_env_or_skip() -> None:
             pytest.skip("No build system found")
 
 
-@pytest.mark.configure
+@pytest.mark.configure()
 def test_cattrs_comparison(tmp_path):
-
     build_dir = tmp_path / "build"
 
     cmake = CMake.default_search(minimum_version=Version("3.15"))
@@ -64,7 +63,7 @@ def test_no_index(tmp_path):
         load_reply_dir_cattrs(tmp_path)
 
 
-@pytest.mark.configure
+@pytest.mark.configure()
 def test_simple_pure(tmp_path):
     build_dir = tmp_path / "build"
 
